"""Bot stats plugin"""
# Copyright (C) 2020 - 2022  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import ClassVar, Optional

from anjani import command, filters, plugin, util


class Topics(plugin.Plugin):
    name: ClassVar[str] = "Topic"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("CHATS")

    @command.filters(filters.can_manage_topic, aliases=["setdefaulttopic"])
    async def cmd_setactiontopic(self, ctx: command.Context) -> Optional[str]:
        """Set action topic"""
        if not ctx.chat.is_forum:
            return await self.text(ctx.chat.id, "topic-non-topic")

        await self.db.update_one(
            {"chat_id": ctx.chat.id},
            {"$set": {"action_topic": ctx.msg.message_thread_id}},
            upsert=True,
        )
        return await self.text(ctx.chat.id, "topic-set")

    # TODO: Add command to create, delete, edit topic



"""Anjani custom types"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import abstractmethod, abstractproperty
from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar

from aiohttp import ClientSession
from pyrogram.filters import Filter

if TYPE_CHECKING:
    from anjani.core import Anjani

Bot = TypeVar("Bot", bound="Anjani", covariant=True)
ChatId = TypeVar("ChatId", int, None, covariant=True)
TextName = TypeVar("TextName", bound=str, covariant=True)
NoFormat = TypeVar("NoFormat", bound=bool, covariant=True)
TypeData = TypeVar("TypeData", covariant=True)
DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., Any])


class Instantiable(Protocol):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError


class CustomFilter(Filter):  # skipcq: PYL-W0223
    anjani: "Anjani"
    include_bot: bool


class NDArray(Protocol[TypeData]):
    @abstractmethod
    def __getitem__(self, key: int) -> Any:
        raise NotImplementedError

    @abstractproperty
    def size(self) -> int:
        raise NotImplementedError


class Classifier(Protocol):
    @abstractmethod
    async def predict(self, text: str, **predict_params: Any) -> NDArray[Any]:
        raise NotImplementedError

    @abstractmethod
    async def load_model(self, http_client: ClientSession) -> None:
        raise NotImplementedError

    @abstractmethod
    async def is_spam(self, text: str) -> bool:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def normalize(text: str) -> str:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def prob_to_string(value: float) -> str:
        raise NotImplementedError


class WebServer(Protocol):
    @abstractmethod
    async def run(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def stop(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def add_router(self, **router_param: Any) -> None:
        raise NotImplementedError


class Router(Instantiable):
    def get(self, *args, **kwargs) -> Callable[[DecoratedCallable], DecoratedCallable]:
        raise NotImplementedError

    def post(self, *args, **kwargs) -> Callable[[DecoratedCallable], DecoratedCallable]:
        raise NotImplementedError

    def put(self, *args, **kwargs) -> Callable[[DecoratedCallable], DecoratedCallable]:
        raise NotImplementedError



"""miscellaneous bot commands"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from json import JSONDecodeError
from typing import Any, ClassVar, Optional

from aiohttp import ClientConnectorError, ClientSession, ContentTypeError
from aiopath import AsyncPath

from anjani import command, filters, plugin


class Paste:
    def __init__(self, session: ClientSession, name: str, url: str):
        self.__session = session
        self.__name = name
        self.__url = url

        self.url_map = {
            "-h": "https://hastebin.com/",
            "-s": "https://stashbin.xyz/",
            "hastebin": "https://hastebin.com/",
            "stashbin": "https://stashbin.xyz/",
            "spacebin": "https://spaceb.in/",
        }

    async def __aenter__(self) -> "Paste":
        return self

    async def __aexit__(self, _: Any, __: Any, ___: Any) -> None:
        ...

    async def go(self, content: Any) -> str:
        async with self.__session.post(self.__url, data=content) as r:
            content_data = await r.json()
            url = self.url_map[self.__name]
            if self.__name == "stashbin":
                slug = content_data["data"]["key"]
            elif self.__name == "hastebin":
                slug = content_data["key"]
            else:
                slug = content_data["payload"]["id"]
            return url + slug


class Misc(plugin.Plugin):
    name: ClassVar[str] = "Miscs"
    helpable: ClassVar[bool] = True

    async def cmd_id(self, ctx: command.Context) -> str:
        """Display ID's"""
        msg = ctx.msg.reply_to_message or ctx.msg
        out_str = f"👥 **Chat ID :** `{(msg.forward_from_chat or msg.chat).id}`\n"
        if msg.is_topic_message:
            out_str += f"🗨️ **Topic ID :** `{msg.topics.id}`\n"
        out_str += f"💬 **Message ID :** `{msg.forward_from_message_id or msg.id}`\n"
        if msg.from_user:
            out_str += f"🙋‍♂️ **From User ID :** `{msg.from_user.id}`\n"
        file = (
            msg.audio
            or msg.animation
            or msg.document
            or msg.photo
            or msg.sticker
            or msg.voice
            or msg.video_note
            or msg.video
        ) or None
        if file:
            out_str += f"📄 **Media Type :** `{file.__class__.__name__}`\n"
            out_str += f"📄 **File ID :** `{file.file_id}`"

        return out_str

    async def cmd_paste(self, ctx: command.Context, service: Optional[str] = None) -> Optional[str]:
        if not ctx.msg.reply_to_message:
            return None

        if not service:
            service = "stashbin"

        chat = ctx.chat
        reply_msg = ctx.msg.reply_to_message

        data: Any
        if reply_msg.document:
            file = AsyncPath(await reply_msg.download())
            data = await file.read_text()
            await file.unlink()
        elif reply_msg.text:
            data = reply_msg.text
        else:
            return None

        uris = {
            "-h": "https://hastebin.com/documents",
            "-s": "http://stashbin.xyz/api/document",
            "hastebin": "https://hastebin.com/documents",
            "stashbin": "http://stashbin.xyz/api/document",
            "spacebin": "https://spaceb.in/api/v1/documents/",
        }
        try:
            uri = uris[service]
        except KeyError:
            return await self.get_text(chat.id, "paste-invalid", service)

        if service in {"-h", "hastebin"}:
            service = "hastebin"
            data = data.encode("utf-8")
        elif service in {"-s", "stashbin"}:
            service = "stashbin"
            data = {"content": data}
        elif service == "spacebin":
            service = "spacebin"
            data = {"content": data, "extension": "txt"}
        else:
            return await self.get_text(chat.id, "paste-invalid", service)

        await ctx.respond(await self.text(chat.id, "paste-wait", service))

        try:
            async with Paste(self.bot.http, service, uri) as paste:
                return await self.text(
                    ctx.chat.id, "paste-succes", f"[{service}]({await paste.go(data)})"
                )
        except (JSONDecodeError, ContentTypeError, ClientConnectorError, KeyError):
            self.log.error("Error while pasting", exc_info=True)
            return await self.text(ctx.chat.id, "paste-fail", service)

    @command.filters(filters.private)
    async def cmd_source(self, ctx: command.Context) -> None:
        """Send the bot source code"""
        await ctx.respond(
            "[GitHub repo](https://github.com/userbotindo/Anjani)\n"
            + "[Support](https://t.me/userbotindo)",
            disable_web_page_preview=True,
        )

    @command.filters(filters.group)
    async def cmd_slap(self, ctx: command.Context) -> Optional[str]:
        """Slap member with neko slap."""
        text = ctx.input
        chat = ctx.msg.chat
        async with self.bot.http.get("https://www.nekos.life/api/v2/img/slap") as slap:
            if slap.status != 200:
                return await self.text(chat.id, "err-api-down")
            res = await slap.json()

        msg = ctx.msg.reply_to_message or ctx.msg
        await self.bot.client.send_animation(
            chat.id,
            res["url"],
            reply_to_message_id=msg.id,
            caption=text,
        )
        return None



""" Canonical plugin for @dAnjani_bot """
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from base64 import b64encode
from datetime import datetime, timezone
from typing import Any, ClassVar, MutableMapping

from aiopath import AsyncPath
from pymongo import UpdateOne
from pymongo.errors import PyMongoError
from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.enums.chat_members_filter import ChatMembersFilter
from pyrogram.enums.message_media_type import MessageMediaType
from pyrogram.types import (
    ChatMemberUpdated,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    LoginUrl,
    Message,
)

try:
    from userbotindo import WebServer

    _run_canonical = True
except ImportError:
    from anjani.util.types import WebServer

    _run_canonical = False


from anjani import command, filters, listener, plugin, util


class Canonical(plugin.Plugin):
    """Helper Plugin
    This plugin is only available for @dAnjani_bot
    to comunicate with https://userbotindo.com
    """

    name: ClassVar[str] = "Canonical"
    disabled: ClassVar[bool] = not _run_canonical

    # Private
    _api: WebServer
    __task: asyncio.Task[None]
    __web_server: asyncio.Task[None]
    _mt: MutableMapping[MessageMediaType, str] = {
        MessageMediaType.STICKER: "s",
        MessageMediaType.PHOTO: "p",
        MessageMediaType.DOCUMENT: "f",
        MessageMediaType.VIDEO: "v",
    }

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("TEST")
        self.db_analytics = self.bot.db.get_collection("ANALYTICS")
        self.chats_db = self.bot.db.get_collection("CHATS")
        self._api = WebServer(
            title="Anjani API Docs", description="API Documentation for Anjani Services"
        )

    async def on_start(self, _: int) -> None:
        self.log.debug("Starting watch streams")
        self.__task = self.bot.loop.create_task(self.watch_streams())

        def server_done_cb(task: asyncio.Task[None]):
            if task.cancelled:
                self.log.debug("Stopping web server")
                asyncio.ensure_future(self._api.stop())

        self.log.debug("Starting web server")
        self.__web_server = self.bot.loop.create_task(self._api.run())
        self.__web_server.add_done_callback(server_done_cb)

    async def on_stop(self) -> None:
        self.log.debug("Stopping watch streams")
        self.__task.cancel()
        self.__web_server.cancel()

    def get_type(self, message: Message) -> str:
        if message.command:
            return "c"
        return self._mt.get(message.media, "t") if message.media else "t"

    async def save_message_type(self, message: Message) -> None:
        today = util.time.sec()
        timestamp = today - (today % 86400)  # truncate to day

        # TODO: Remove old schema after analytics migration
        await self.db_analytics.bulk_write(
            [
                UpdateOne(
                    {"key": 2},
                    {"$inc": {f"data.{str(timestamp)}.{self.get_type(message)}": 1}},
                    upsert=True,
                ),
                UpdateOne(  # new schema
                    {
                        "timestamp": datetime.now(timezone.utc).replace(
                            minute=0, second=0, microsecond=0
                        )
                    },
                    {"$inc": {f"data.{self.get_type(message)}": 1}},
                    upsert=True,
                ),
            ]
        )

    @command.filters(filters.admin_only & filters.group)
    async def cmd_r(self, ctx: command.Context) -> None:
        """Refresh chat data"""
        admins = []
        async for member in self.bot.client.get_chat_members(
            ctx.chat.id, filter=ChatMembersFilter.ADMINISTRATORS
        ):  # type: ignore
            if not member.user.is_bot and member.privileges.can_manage_chat:
                admins.append(member.user.id)

        photo = None
        if ctx.chat.photo:
            file = await self.bot.client.download_media(ctx.chat.photo.small_file_id)
            if file and isinstance(file, str):
                file = AsyncPath(file)
                result = await file.read_bytes()
                photo = b64encode(result).decode("utf-8")
                await file.unlink()
            if file and isinstance(file, bytes):
                photo = b64encode(file).decode("utf-8")

        await self.chats_db.update_one(
            {"chat_id": ctx.chat.id}, {"$set": {"admins": admins, "photo": photo}}, upsert=True
        )
        await ctx.respond("Done", delete_after=5)

    @listener.priority(65)
    async def on_message(self, message: Message) -> None:
        if message.outgoing:
            return

        # Analytics
        self.bot.loop.create_task(self.save_message_type(message))

    async def on_chat_action(self, message: Message) -> None:
        """Delete admins data from chats"""
        if message.new_chat_members:
            return

        chat = message.chat
        user = message.left_chat_member
        if user.id == self.bot.uid:
            await asyncio.gather(
                self.chats_db.update_one({"chat_id": chat.id}, {"$set": {"admins": []}}),
            )

    async def on_chat_member_update(self, update: ChatMemberUpdated) -> None:
        old_data = update.old_chat_member
        new_data = update.new_chat_member

        if not old_data or not new_data:  # Rare case
            return

        if old_data.status == new_data.status:
            return

        if new_data.status not in {
            ChatMemberStatus.ADMINISTRATOR,
            ChatMemberStatus.OWNER,
        }:
            await self.chats_db.update_one(
                {"chat_id": update.chat.id}, {"$pull": {"admins": new_data.user.id}}
            )
        elif (
            new_data.status in {ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER}
            and new_data.privileges
            and new_data.privileges.can_manage_chat
        ):  # type: ignore
            await self.chats_db.update_one(
                {"chat_id": update.chat.id},
                {"$addToSet": {"admins": new_data.user.id}},
                upsert=True,
            )

    async def watch_streams(self) -> None:
        try:
            async with self.db.watch([{"$match": {"operationType": "insert"}}]) as stream:
                async for change in stream:
                    await self.dispatch_change(change["fullDocument"])
        except PyMongoError as e:
            self.log.error("Error", exc_info=e)

    async def dispatch_change(self, doc: MutableMapping[str, Any]) -> None:
        chat_id = int(doc["_id"])
        message = doc["message"]
        pin = doc.get("pin", False)
        disable_preview = doc.get("disable_preview", False)

        try:
            msg = await self.bot.client.send_message(
                chat_id=chat_id,
                text=message,
                disable_web_page_preview=disable_preview,
            )
            if pin:
                await self.bot.client.pin_chat_message(msg.chat.id, msg.id)
        except Exception as e:  # skipcq: PYL-W0703
            self.log.error(f"Error sending message to {chat_id}", exc_info=e)
        finally:
            await self.db.delete_one({"_id": chat_id})

    @command.filters(filters.private)
    async def cmd_login(self, ctx: command.Context):
        """Login to https://userbotindo.com"""
        if not self.bot.config.LOGIN_URL:
            return

        await ctx.respond(
            "Click this button to login to Anjani Dashboard",
            reply_markup=InlineKeyboardMarkup(
                [
                    [
                        InlineKeyboardButton(
                            "Login",
                            login_url=LoginUrl(
                                url=self.bot.config.LOGIN_URL,
                                forward_text="Login to https://userbotindo.com",
                                request_write_access="True",
                            ),
                        )
                    ]
                ]
            ),
        )



"""Anjani main entry point"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging
import os
import sys
from pathlib import Path
from typing import Any, MutableMapping

import aiorun
import colorlog
import dotenv

from . import DEFAULT_CONFIG_PATH
from .core import Anjani
from .util.config import Config

log = logging.getLogger("launch")


def _level_check(level: str) -> int:
    _str_to_lvl = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
    }
    if level not in _str_to_lvl:
        return logging.INFO

    return _str_to_lvl[level]


def _setup_log() -> None:
    """Configures logging"""
    level = _level_check(os.environ.get("LOG_LEVEL", "info").upper())
    logging.root.setLevel(level)

    # Color log config
    log_color: bool = os.environ.get("LOG_COLOR") in {"enable", 1, "1", "true"}

    file_format = "[ %(asctime)s: %(levelname)-8s ] %(name)-15s - %(message)s"
    logfile = logging.FileHandler("Anjani.log")
    formatter = logging.Formatter(file_format, datefmt="%H:%M:%S")
    logfile.setFormatter(formatter)
    logfile.setLevel(level)

    if log_color:
        formatter = colorlog.ColoredFormatter(
            "  %(log_color)s%(levelname)-8s%(reset)s  |  "
            "%(name)-15s  |  %(log_color)s%(message)s%(reset)s"
        )
    else:
        formatter = logging.Formatter("  %(levelname)-8s  |  %(name)-15s  |  %(message)s")
    stream = logging.StreamHandler()
    stream.setLevel(level)
    stream.setFormatter(formatter)

    root = logging.getLogger()
    root.setLevel(level)
    root.addHandler(stream)
    root.addHandler(logfile)

    # Logging necessary for selected libs
    aiorun.logger.disabled = True
    logging.getLogger("pymongo").setLevel(logging.WARNING)
    logging.getLogger("pyrogram").setLevel(logging.ERROR)
    logging.getLogger("urllib3").setLevel(logging.WARNING)


def start() -> None:
    """Main entry point for the bot."""
    config_path = Path(DEFAULT_CONFIG_PATH)
    if config_path.is_file():
        dotenv.load_dotenv(config_path)

    _setup_log()
    log.info(
        "Running on Python %s.%s.%s",
        sys.version_info.major,
        sys.version_info.minor,
        sys.version_info.micro,
    )
    log.info("Loading code")

    _uvloop = False
    if sys.platform == "win32":
        policy = asyncio.WindowsProactorEventLoopPolicy()
        asyncio.set_event_loop_policy(policy)
    else:
        try:
            import uvloop  # type: ignore
        except ImportError:
            pass
        else:
            uvloop.install()
            _uvloop = True
            log.info("Using uvloop event loop")

    log.info("Initializing bot")
    loop = asyncio.new_event_loop()

    aiorun.run(Anjani.init_and_run(Config(), loop=loop), loop=loop if _uvloop else None)



"""Anjani async utils"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import functools
from typing import Any, Callable, TypeVar

Result = TypeVar("Result")


async def run_sync(func: Callable[..., Result], *args: Any, **kwargs: Any) -> Result:
    """Runs the given sync function (optionally with arguments) on a separate thread."""

    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, functools.partial(func, *args, **kwargs))



"""Anjani database cursor"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Deque,
    Generic,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from bson.code import Code
from pymongo.cursor import Cursor as _Cursor
from pymongo.typings import _CollationIn, _DocumentType

from anjani import util

from .cursor_base import AsyncCursorBase

if TYPE_CHECKING:
    from .collection import AsyncCollection


class Cursor(_Cursor, Generic[_DocumentType]):

    _Cursor__data: Deque[Any]
    _Cursor__killed: bool
    _Cursor__query_flags: int

    delegate: "AsyncCollection[_DocumentType]"

    def __init__(
        self, collection: "AsyncCollection[_DocumentType]", *args: Any, **kwargs: Any
    ) -> None:
        self.delegate = collection

        super().__init__(collection.dispatch, *args, **kwargs)

    @abstractmethod
    def add_option(self, mask: int) -> "Cursor[_DocumentType]":
        raise NotImplementedError

    @abstractmethod
    def allow_disk_use(self, allow_disk_use: bool) -> "Cursor[_DocumentType]":
        raise NotImplementedError

    @abstractmethod
    def collation(self, collation: Optional[_CollationIn]) -> "Cursor[_DocumentType]":
        raise NotImplementedError

    @abstractmethod
    def comment(self, comment: str) -> "Cursor[_DocumentType]":
        raise NotImplementedError

    @property
    def _AsyncCursor__data(self) -> Deque[Any]:
        return self.__data

    async def _AsyncCursor__die(self, synchronous: bool = False) -> None:
        await util.run_sync(self.__die, synchronous=synchronous)

    @property
    def _AsyncCursor__exhaust(self) -> bool:
        return self.__exhaust

    @property
    def _AsyncCursor__killed(self) -> bool:
        return self.__killed

    @property
    def _AsyncCursor__max_await_time_ms(self) -> Optional[int]:
        return self.__max_await_time_ms

    @property
    def _AsyncCursor__max_time_ms(self) -> Optional[int]:
        return self.__max_time_ms

    @property
    def _AsyncCursor__query_flags(self) -> int:
        return self.__query_flags

    @property
    def _AsyncCursor__query_spec(self) -> Optional[Any]:
        return self.__query_spec

    @property
    def _AsyncCursor__retrieved(self) -> int:
        return self.__retrieved

    @property
    def _AsyncCursor__spec(self) -> Mapping[str, Any]:
        return self.__spec

    @property
    def collection(self) -> "AsyncCollection[_DocumentType]":
        return self.delegate


class RawBatchCursor(Cursor):  # skipcq: PYL-W0223
    pass


class AsyncCursor(AsyncCursorBase, Generic[_DocumentType]):
    """AsyncIO :obj:`~Cursor`

    *DEPRECATED* methods are removed in this class.
    """

    dispatch: _Cursor

    def add_option(self, mask: int) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.add_option(mask)
        return self

    def allow_disk_use(self, allow_disk_use: bool) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.allow_disk_use(allow_disk_use)
        return self

    def collation(self, collation: _CollationIn) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.collation(collation)
        return self

    def comment(self, comment: str) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.comment(comment)
        return self

    async def distinct(self, key: str) -> List[Any]:
        return await util.run_sync(self.dispatch.distinct, key)

    async def explain(self) -> _DocumentType:
        return await util.run_sync(self.dispatch.explain)

    def hint(self, index: Union[str, List[Tuple[str, Any]]]) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.hint(index)
        return self

    def limit(self, limit: int) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.limit(limit)
        return self

    def max(self, spec: List[Any]) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.max(spec)
        return self

    def max_await_time_ms(self, max_await_time_ms: int) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.max_await_time_ms(max_await_time_ms)
        return self

    def max_time_ms(self, max_time_ms: int) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.max_time_ms(max_time_ms)
        return self

    def min(self, spec: List[Any]) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.min(spec)
        return self

    def remove_option(self, mask: int) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.remove_option(mask)
        return self

    def rewind(self) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.rewind()
        return self

    def skip(self, skip: int) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.skip(skip)
        return self

    def sort(
        self, key: Union[str, List[Tuple[str, Any]]], *, direction: Any = None
    ) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.sort(key, direction=direction)
        return self

    def where(self, code: Code) -> "AsyncCursor[_DocumentType]":
        self.dispatch = self.dispatch.where(code)
        return self

    def _query_flags(self) -> int:
        # skipcq: PYL-W0212
        return self.dispatch._Cursor__query_flags  # type: ignore

    def _data(self) -> Deque[Any]:
        # skipcq: PYL-W0212
        return self.dispatch._Cursor__data  # type: ignore

    def _killed(self) -> bool:
        # skipcq: PYL-W0212
        return self.dispatch._Cursor__killed  # type: ignore


class AsyncRawBatchCursor(AsyncCursor, Generic[_DocumentType]):
    pass



"""Anjani database client"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from contextlib import asynccontextmanager
from typing import (
    Any,
    AsyncGenerator,
    FrozenSet,
    List,
    Literal,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
)

from bson.codec_options import DEFAULT_CODEC_OPTIONS, CodecOptions
from bson.son import SON
from bson.timestamp import Timestamp
from pymongo.client_session import TransactionOptions
from pymongo.collation import Collation
from pymongo.driver_info import DriverInfo
from pymongo.mongo_client import MongoClient
from pymongo.read_concern import ReadConcern
from pymongo.read_preferences import ReadPreference
from pymongo.topology_description import TopologyDescription
from pymongo.typings import _Address
from pymongo.write_concern import DEFAULT_WRITE_CONCERN, WriteConcern

from anjani import util

from .base import AsyncBaseProperty
from .change_stream import AsyncChangeStream
from .client_session import AsyncClientSession
from .command_cursor import AsyncCommandCursor, CommandCursor
from .db import AsyncDatabase
from .typings import ReadPreferences


class AsyncClient(AsyncBaseProperty):
    """AsyncIO :obj:`~MongoClient`

    *DEPRECATED* methods are removed in this class.
    """

    dispatch: MongoClient

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs.update(
            {"driver": DriverInfo("AsyncIOMongoDB", version="staging", platform="AsyncIO")}
        )
        dispatch = MongoClient(*args, **kwargs)

        # Propagate initialization to base
        super().__init__(dispatch)

    def __getitem__(self, name: str) -> AsyncDatabase:
        return AsyncDatabase(self, self.dispatch[name])

    def __hash__(self) -> int:
        return hash(self.address)

    async def close(self) -> None:
        await util.run_sync(self.dispatch.close)

    async def drop_database(
        self,
        name_or_database: Union[str, AsyncDatabase],
        session: Optional[AsyncClientSession] = None,
    ) -> None:
        if isinstance(name_or_database, AsyncDatabase):
            name_or_database = name_or_database.name

        return await util.run_sync(
            self.dispatch.drop_database,
            name_or_database,
            session=session.dispatch if session else session,
        )

    def get_database(
        self,
        name: Optional[str] = None,
        *,
        codec_options: Optional[CodecOptions] = None,
        read_preference: Optional[ReadPreferences] = None,
        write_concern: Optional[WriteConcern] = None,
        read_concern: Optional[ReadConcern] = None,
    ) -> AsyncDatabase:
        return AsyncDatabase(
            self,
            self.dispatch.get_database(
                name,
                codec_options=codec_options,
                read_preference=read_preference,
                write_concern=write_concern,
                read_concern=read_concern,
            ),
        )

    def get_default_database(
        self,
        default: Optional[str] = None,
        *,
        codec_options: Optional[CodecOptions] = None,
        read_preference: Optional[ReadPreferences] = None,
        write_concern: Optional[WriteConcern] = None,
        read_concern: Optional[ReadConcern] = None,
    ) -> AsyncDatabase:
        return AsyncDatabase(
            self,
            self.dispatch.get_default_database(
                default,
                codec_options=codec_options,
                read_preference=read_preference,
                write_concern=write_concern,
                read_concern=read_concern,
            ),
        )

    async def list_database_names(self, session: Optional[AsyncClientSession] = None) -> List[str]:
        return await util.run_sync(
            self.dispatch.list_database_names, session=session.dispatch if session else session
        )

    async def list_databases(
        self, session: Optional[AsyncClientSession] = None, **kwargs: Any
    ) -> AsyncCommandCursor:
        cmd = SON([("listDatabases", 1)])
        cmd.update(kwargs)
        database = self.get_database(
            "admin",
            codec_options=DEFAULT_CODEC_OPTIONS,
            read_preference=ReadPreference.PRIMARY,
            write_concern=DEFAULT_WRITE_CONCERN,
        )
        res: Mapping[str, Any] = await util.run_sync(
            database.dispatch._retryable_read_command,  # skipcq: PYL-W0212
            cmd,
            session=session.dispatch if session else session,
        )
        cursor: Mapping[str, Any] = {
            "id": 0,
            "firstBatch": res["databases"],
            "ns": "admin.$cmd",
        }
        return AsyncCommandCursor(CommandCursor(database["$cmd"], cursor, None))

    async def server_info(self, session: Optional[AsyncClientSession] = None) -> Mapping[str, Any]:
        return await util.run_sync(
            self.dispatch.server_info, session=session.dispatch if session else session
        )

    # Don't need await when entering the context manager,
    # because it's slightly different than motor libs.
    @asynccontextmanager
    async def start_session(
        self,
        *,
        causal_consistency: Optional[bool] = None,
        default_transaction_options: Optional[TransactionOptions] = None,
        snapshot: bool = False,
    ) -> AsyncGenerator[AsyncClientSession, None]:
        session = await util.run_sync(
            self.dispatch.start_session,
            causal_consistency=causal_consistency,
            default_transaction_options=default_transaction_options,
            snapshot=snapshot,
        )

        async with AsyncClientSession(self, session) as session:
            yield session

    def watch(
        self,
        pipeline: Optional[List[Mapping[str, Any]]] = None,
        *,
        full_document: Optional[Literal["updateLookup"]] = None,
        resume_after: Optional[Mapping[str, str]] = None,
        max_await_time_ms: Optional[int] = None,
        batch_size: Optional[int] = None,
        collation: Optional[Collation] = None,
        start_at_operation_time: Optional[Timestamp] = None,
        session: Optional[AsyncClientSession] = None,
        start_after: Optional[Mapping[str, str]] = None,
        comment: Optional[str] = None,
        full_document_before_change: Optional[Literal["required", "whenAvailable"]] = None,
    ) -> AsyncChangeStream:
        return AsyncChangeStream(
            self,
            pipeline,
            full_document,
            resume_after,
            max_await_time_ms,
            batch_size,
            collation,
            start_at_operation_time,
            session,
            start_after,
            comment,
            full_document_before_change,
        )

    @property
    def HOST(self) -> str:
        return self.dispatch.HOST

    @property
    def PORT(self) -> int:
        return self.dispatch.PORT

    @property
    def address(self) -> Optional[Tuple[str, int]]:
        return self.dispatch.address

    @property
    def arbiters(self) -> Set[Tuple[str, int]]:
        return self.dispatch.arbiters

    @property
    def event_listeners(self) -> Any:
        return self.dispatch.event_listeners

    @property
    def is_mongos(self) -> bool:
        return self.dispatch.is_mongos

    @property
    def is_primary(self) -> bool:
        return self.dispatch.is_primary

    @property
    def nodes(self) -> FrozenSet[_Address]:
        return self.dispatch.nodes

    @property
    def primary(self) -> Optional[Tuple[str, int]]:
        return self.dispatch.primary

    @property
    def secondaries(self) -> Set[Tuple[str, int]]:
        return self.dispatch.secondaries

    @property
    def topology_description(self) -> TopologyDescription:
        return self.dispatch.topology_description



"""Anjani time utils"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time as t  # skipcq: PYL-W0406
from typing import Union


def usec() -> int:
    """Returns the current time in microseconds since the Unix epoch."""

    return int(t.time() * 1000000)


def msec() -> int:
    """Returns the current time in milliseconds since the Unix epoch."""

    return int(usec() / 1000)


def sec() -> int:
    """Returns the current time in seconds since the Unix epoch."""

    return int(t.time())


def format_duration_us(t_us: Union[int, float]) -> str:
    """Formats the given microsecond duration as a string."""
    t_us = int(t_us)

    t_ms = t_us / 1000
    t_s = t_ms / 1000
    t_m = t_s / 60
    t_h = t_m / 60
    t_d = t_h / 24

    if t_d >= 1:
        rem_h = t_h % 24
        return f"{int(t_d)}d {int(rem_h)}h"

    if t_h >= 1:
        rem_m = t_m % 60
        return f"{int(t_h)}h {int(rem_m)}m"

    if t_m >= 1:
        rem_s = t_s % 60
        return f"{int(t_m)}m {int(rem_s)}s"

    if t_s >= 1:
        return f"{int(t_s)}s"

    if t_ms >= 1:
        return f"{int(t_ms)}ms"

    return f"{int(t_us)}us"


def extract_time(time_text: str) -> Union[int, bool]:
    """Extract time from time flags"""
    if any(time_text.endswith(unit) for unit in ("m", "h", "d")):
        unit = time_text[-1]
        time_num = time_text[:-1]
        if not time_num.isdigit():
            return False

        if unit == "m":
            return int(t.time() + int(time_num) * 60)

        if unit == "h":
            return int(t.time() + int(time_num) * 60 * 60)

        if unit == "d":
            return int(t.time() + int(time_num) * 24 * 60 * 60)

    return False



"""Anjani database core"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any, Generic, Union

from bson.codec_options import CodecOptions
from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.command_cursor import CommandCursor, RawBatchCommandCursor
from pymongo.cursor import Cursor, RawBatchCursor
from pymongo.database import Database
from pymongo.mongo_client import MongoClient
from pymongo.read_concern import ReadConcern
from pymongo.read_preferences import _ServerMode
from pymongo.typings import _DocumentType
from pymongo.write_concern import WriteConcern

if TYPE_CHECKING:
    from .command_cursor import _LatentCursor


class AsyncBase(Generic[_DocumentType]):
    """Base Class for AsyncIOMongoDB Instances"""

    dispatch: Union[
        "_LatentCursor[_DocumentType]",
        ClientSession,
        Collection[_DocumentType],
        CommandCursor[_DocumentType],
        Cursor[_DocumentType],
        Database[_DocumentType],
        MongoClient[_DocumentType],
        RawBatchCursor[_DocumentType],
        RawBatchCommandCursor[_DocumentType],
    ]

    def __init__(
        self,
        dispatch: Union[
            "_LatentCursor[_DocumentType]",
            ClientSession,
            Collection[_DocumentType],
            CommandCursor[_DocumentType],
            Cursor[_DocumentType],
            Database[_DocumentType],
            MongoClient[_DocumentType],
            RawBatchCursor[_DocumentType],
            RawBatchCommandCursor[_DocumentType],
        ],
    ) -> None:
        self.dispatch = dispatch

    def __eq__(self, other: Any) -> bool:
        if (
            isinstance(other, self.__class__)
            and hasattr(self, "dispatch")
            and hasattr(other, "dispatch")
        ):
            return self.dispatch == other.dispatch

        return NotImplemented

    def __hash__(self):
        return hash(self.dispatch)

    def __repr__(self) -> str:
        return type(self).__name__ + f"({self.dispatch!r})"


class AsyncBaseProperty(AsyncBase):
    """Base class property for AsyncIOMongoDB instances"""

    dispatch: Union[Collection, Database, MongoClient]

    @property
    def codec_options(self) -> CodecOptions:
        return self.dispatch.codec_options

    @property
    def read_preference(self) -> _ServerMode:
        return self.dispatch.read_preference

    @property
    def read_concern(self) -> ReadConcern:
        return self.dispatch.read_concern

    @property
    def write_concern(self) -> WriteConcern:
        return self.dispatch.write_concern



"""Anjani database commmand cursor"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from collections import deque
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Coroutine,
    Deque,
    Generic,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from pymongo.client_session import ClientSession
from pymongo.command_cursor import CommandCursor as _CommandCursor
from pymongo.typings import _Address, _DocumentType

from anjani import util

from .client_session import AsyncClientSession
from .cursor_base import AsyncCursorBase

if TYPE_CHECKING:
    from .collection import AsyncCollection


class CommandCursor(_CommandCursor, Generic[_DocumentType]):

    _CommandCursor__data: Deque[Any]
    _CommandCursor__killed: bool

    delegate: "AsyncCollection"

    def __init__(
        self,
        collection: "AsyncCollection",
        cursor_info: Mapping[str, Any],
        address: Optional[Union[Tuple[str, int], _Address]] = None,
        *,
        batch_size: int = 0,
        max_await_time_ms: Optional[int] = None,
        session: Optional[AsyncClientSession] = None,
        explicit_session: bool = False,
    ) -> None:
        self.delegate = collection

        super().__init__(
            collection.dispatch,
            cursor_info,
            address,
            batch_size=batch_size,
            max_await_time_ms=max_await_time_ms,
            session=session.dispatch if session else session,
            explicit_session=explicit_session,
        )

    async def _AsyncCommandCursor__die(self, synchronous: bool = False) -> None:
        await util.run_sync(self.__die, synchronous=synchronous)

    @property
    def _AsyncCommandCursor__data(self) -> Deque[Any]:
        return self.__data

    @property
    def _AsyncCommandCursor__killed(self) -> bool:
        return self.__killed

    @property
    def collection(self) -> "AsyncCollection[_DocumentType]":
        return self.delegate


class RawBatchCommandCursor(CommandCursor, Generic[_DocumentType]):
    pass


class AsyncCommandCursor(AsyncCursorBase):
    """AsyncIO :obj:`~CommandCursor`

    *DEPRECATED* methods are removed in this class.
    """

    dispatch: CommandCursor

    def _query_flags(self) -> int:
        return 0

    def _data(self) -> Deque[Any]:
        return self.dispatch._CommandCursor__data  # skipcq: PYL-W0212

    def _killed(self) -> bool:
        return self.dispatch._CommandCursor__killed  # skipcq: PYL-W0212


class _LatentCursor(Generic[_DocumentType]):
    """Base class for LatentCursor AsyncIOMongoDB instance"""

    # ClassVar
    alive: ClassVar[bool] = True
    _CommandCursor__data: ClassVar[Deque[Any]] = deque()
    _CommandCursor__id: ClassVar[Optional[Any]] = None
    _CommandCursor__killed: ClassVar[bool] = False
    _CommandCursor__sock_mgr: ClassVar[Optional[Any]] = None
    _CommandCursor__session: ClassVar[Optional[AsyncClientSession]] = None
    _CommandCursor__explicit_session: ClassVar[Optional[bool]] = None
    address: ClassVar[Optional[Union[Tuple[str, int], _Address]]] = None
    cursor_id: ClassVar[Optional[Any]] = None
    session: Optional[ClientSession] = None

    _CommandCursor__collection: "AsyncCollection"

    def __init__(self, collection: "AsyncCollection") -> None:
        self._CommandCursor__collection = collection

    def _CommandCursor__end_session(self, *args: Any, **kwargs: Any) -> None:
        pass  # Only for initialization

    def _CommandCursor__die(self, *args: Any, **kwargs: Any) -> None:
        pass  # Only for initialization

    def _refresh(self) -> int:  # skipcq: PYL-R0201
        """Only for initialization"""
        return 0

    def batch_size(self, batch_size: int) -> None:
        pass  # Only for initialization

    def close(self) -> None:
        pass  # Only for initialization

    def clone(self) -> "_LatentCursor":
        return _LatentCursor(self._CommandCursor__collection)

    def rewind(self):
        pass  # Only for initialization

    @property
    def collection(self):
        return self._CommandCursor__collection


class AsyncLatentCommandCursor(AsyncCommandCursor):
    """Temporary Cursor for initializing in aggregate,
    and will be overwrite by :obj:`~asyncio.Future`"""

    dispatch: Union[CommandCursor, RawBatchCommandCursor]

    def __init__(
        self,
        collection: "AsyncCollection",
        start: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.start = start
        self.args = args
        self.kwargs = kwargs

        super().__init__(_LatentCursor(collection), collection)

    def batch_size(self, batch_size: int) -> "AsyncLatentCommandCursor":
        self.kwargs["batchSize"] = batch_size

        return self

    def _get_more(self) -> Union[asyncio.Future[int], Coroutine[Any, Any, int]]:
        if not self.started:
            self.started = True
            original_future = self.loop.create_future()
            future = self.loop.create_task(util.run_sync(self.start, *self.args, **self.kwargs))
            future.add_done_callback(
                partial(self.loop.call_soon_threadsafe, self._on_started, original_future)
            )

            self.start, self.args, self.kwargs = lambda _: None, (), {}

            return original_future

        return super()._get_more()

    def _on_started(
        self,
        original_future: asyncio.Future[int],
        future: asyncio.Future[Union[CommandCursor, RawBatchCommandCursor]],
    ) -> None:
        try:
            self.dispatch = future.result()
        except Exception as exc:  # skipcq: PYL-W0703
            if not original_future.done():
                original_future.set_exception(exc)
        else:
            # Return early if the task was cancelled.
            if original_future.done():
                return

            if self.dispatch._CommandCursor__data or not self.dispatch.alive:  # skipcq: PYL-W0212
                # _get_more is complete.
                original_future.set_result(
                    len(self.dispatch._CommandCursor__data)  # skipcq: PYL-W0212
                )
            else:
                # Send a getMore.
                fut = self.loop.create_task(super()._get_more())

                def copy(f: asyncio.Future[int]) -> None:
                    if original_future.done():
                        return

                    exc = f.exception()
                    if exc is not None:
                        original_future.set_exception(exc)
                    else:
                        original_future.set_result(f.result())

                fut.add_done_callback(copy)
